unit AppLauncher;

// AppLauncher 1.0.4
// Delphi 2009/2010/XE1/XE2 (Win32 only).

// ISC LICENSE (ISCL)

// Copyright (c) 2009-2012, Dieter Khler based on code by Apollonius
//
// Permission to use, copy, modify, and/or distribute this software for any
// purpose with or without fee is hereby granted, provided that the above
// copyright notice and this permission notice appear in all copies.
//
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
// WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
// ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
// ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
// OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

// HISTORY
//
// 2012-02-09 1.0.4 Support for Windows 98, 95, and ME canceled.
// 2010-09-11 1.0.3 Bug fix.
// 2010-01-22 1.0.2 Minor bug fix.
// 2009-01-31 1.0.1 Functions added to catch standard error output separately.
// 2009-01-29 1.0.0

interface

uses
  Windows, Classes, SysUtils;

type

  TAppLauncherProgressEvent = procedure(Sender: TObject; ProcessInfo: PROCESS_INFORMATION; Position: Int64) of object;

  TAppLauncher = class(TComponent)
  private
    FCreateNoWindow: Boolean;
    FOnStdErrProgress: TAppLauncherProgressEvent;
    FOnStdOutProgress: TAppLauncherProgressEvent;
    FSecurityAttributes: SECURITY_ATTRIBUTES;
    FSecurityDescriptor: SECURITY_DESCRIPTOR;
    function CentralLoop(const ProcessInfo: PROCESS_INFORMATION; const ReadHandle: THandle; const OutputStream: TStream): DWORD; inline;
    function CentralLoopEx(const ProcessInfo: PROCESS_INFORMATION; const ReadHandle, ErrorReadHandle: THandle; const OutputStream, ErrorStream: TStream): DWORD; inline;
  protected
    procedure DoErrorProgress(ProcessInfo: PROCESS_INFORMATION; Position: Int64); virtual;
    procedure DoOutputProgress(ProcessInfo: PROCESS_INFORMATION; Position: Int64); virtual;
    procedure PreparePipe(out ReadHandle, WriteHandle: THandle); virtual;
    function PrepareStartupInfoA(const StdOutput: THandle): STARTUPINFOA; overload; virtual;
    function PrepareStartupInfoA(const StdOutput, StdError: THandle): STARTUPINFOA; overload; virtual;
    function PrepareStartupInfoW(const StdOutput: THandle): STARTUPINFOW; overload; virtual;
    function PrepareStartupInfoW(const StdOutput, StdError: THandle): STARTUPINFOW; overload; virtual;

    property SecurityAttributes: SECURITY_ATTRIBUTES read FSecurityAttributes;
    property SecurityDescriptor: SECURITY_DESCRIPTOR read FSecurityDescriptor;
  public
    constructor Create(AOwner: TComponent); override;
    function ExecuteA(const Path, CmdLine: AnsiString; out StdOut: TBytes): DWORD; overload; virtual;
    function ExecuteA(const Path, CmdLine: AnsiString; out StdOut, StdErr: TBytes): DWORD; overload; virtual;
    function ExecuteW(const Path, CmdLine: WideString; out StdOut: TBytes): DWORD; overload; virtual;
    function ExecuteW(const Path, CmdLine: WideString; out StdOut, StdErr: TBytes): DWORD; overload; virtual;
  published
    property CreateNoWindow: Boolean read FCreateNoWindow write FCreateNoWindow default false;
    property OnStdErrProgress: TAppLauncherProgressEvent read FOnStdErrProgress write FOnStdErrProgress;
    property OnStdOutProgress: TAppLauncherProgressEvent read FOnStdOutProgress write FOnStdOutProgress;
  end;

procedure Register;

implementation

uses
  Forms;

procedure Register;
begin
  RegisterComponents('Open XML', [TAppLauncher]);
end;

{ TAppLauncher }

constructor TAppLauncher.Create(AOwner: TComponent);
begin
  inherited;

  FCreateNoWindow := false;

  ZeroMemory(@FSecurityAttributes, SizeOf(FSecurityAttributes));

  InitializeSecurityDescriptor(@FSecurityDescriptor, SECURITY_DESCRIPTOR_REVISION);
  SetSecurityDescriptorDacl(@FSecurityDescriptor, True, nil, False);
  FSecurityAttributes.lpSecurityDescriptor := @FSecurityDescriptor;
  FSecurityAttributes.nLength := SizeOf(SECURITY_ATTRIBUTES);
  FSecurityAttributes.bInheritHandle := True;
end;

procedure TAppLauncher.DoErrorProgress(ProcessInfo: PROCESS_INFORMATION; Position: Int64);
begin
  if Assigned(FOnStdErrProgress) then
    FOnStdErrProgress(Self, ProcessInfo, Position);
end;

procedure TAppLauncher.DoOutputProgress(ProcessInfo: PROCESS_INFORMATION; Position: Int64);
begin
  if Assigned(FOnStdOutProgress) then
    FOnStdOutProgress(Self, ProcessInfo, Position);
end;

procedure TAppLauncher.PreparePipe(out ReadHandle,
  WriteHandle: THandle);
var
  PipeName: string;
begin
  PipeName := '\\.\pipe\' + IntToHex(Random(MaxInt), 8) + IntToHex(GetCurrentProcessId, 8) + IntToHex(Random(MaxInt), 8);

  ReadHandle := CreateNamedPipe(PChar(PipeName), PIPE_ACCESS_INBOUND or FILE_FLAG_OVERLAPPED, 0, 1, 1024, 1024, 0, nil);
  if ReadHandle = INVALID_HANDLE_VALUE then
    RaiseLastOSError;
  try
    WriteHandle := CreateFile(PChar(PipeName),
                              GENERIC_WRITE,
                              FILE_SHARE_READ,
                              @SecurityAttributes,
                              OPEN_EXISTING,
                              FILE_ATTRIBUTE_NORMAL,
                              0);
    if WriteHandle = INVALID_HANDLE_VALUE then
      RaiseLastOSError;
  except
    CloseHandle(ReadHandle);
  end;
end;

function TAppLauncher.PrepareStartupInfoA(const StdOutput: THandle): STARTUPINFOA;
begin
  Result := PrepareStartupInfoA(StdOutput, StdOutput);
end;

function TAppLauncher.PrepareStartupInfoA(const StdOutput, StdError: THandle): STARTUPINFOA;
begin
  ZeroMemory(@Result, SizeOf(Result));
  Result.cb := SizeOf(Result);
  Result.dwFlags := STARTF_USESTDHANDLES;
  Result.hStdOutput := StdOutput;
  Result.hStdError := StdError;
end;

function TAppLauncher.PrepareStartupInfoW(const StdOutput: THandle): STARTUPINFOW;
begin
  Result := PrepareStartupInfoW(StdOutput, StdOutput);
end;

function TAppLauncher.PrepareStartupInfoW(const StdOutput, StdError: THandle): STARTUPINFOW;
begin
  ZeroMemory(@Result, SizeOf(Result));
  Result.cb := SizeOf(Result);
  Result.dwFlags := STARTF_USESTDHANDLES;
  Result.hStdOutput := StdOutput;
  Result.hStdError := StdError;
end;

function TAppLauncher.CentralLoop(const ProcessInfo: PROCESS_INFORMATION;
  const ReadHandle: THandle; const OutputStream: TStream): DWORD;
const
  BUFFER_SIZE = 512;
var
  Buffer: array[0..BUFFER_SIZE - 1] of Byte;
  BytesRead: Cardinal;
  HandleArray: array[0..1] of THandle;
  OvLapped: OVERLAPPED;
  WaitResult: Cardinal;
begin
  ZeroMemory(@OvLapped, SizeOf(OvLapped));
  OvLapped.hEvent := CreateEvent(nil, True, False, nil);
  try
    HandleArray[0] := ProcessInfo.hProcess;
    HandleArray[1] := OvLapped.hEvent;
    if not ReadFile(ReadHandle, Buffer, BUFFER_SIZE, BytesRead, @OvLapped)
       and (GetLastError <> ERROR_IO_PENDING) then
      RaiseLastOSError;

    while not Application.Terminated do
    begin
      WaitResult := MsgWaitForMultipleObjects(Length(HandleArray), HandleArray[0], False, INFINITE, QS_ALLINPUT);
      if WaitResult = WAIT_OBJECT_0 + 1 then begin

        if not GetOverlappedResult(ReadHandle, OvLapped, BytesRead, False) then
          RaiseLastOSError;
          OutputStream.Write(Buffer, BytesRead);

        ResetEvent(OvLapped.hEvent);
        if not ReadFile(ReadHandle, Buffer, BUFFER_SIZE, BytesRead, @OvLapped)
           and (GetLastError <> ERROR_IO_PENDING) then
          RaiseLastOSError;

        DoOutputProgress(ProcessInfo, OutputStream.Position);

      end else if WaitResult = WAIT_OBJECT_0 + 2 then // Message received.
        Application.ProcessMessages
      else
        break; // Process terminated.

    end;

    GetExitCodeProcess(ProcessInfo.hProcess, Result);
  finally
    CloseHandle(OvLapped.hEvent);
  end;
end;

function TAppLauncher.CentralLoopEx(const ProcessInfo: PROCESS_INFORMATION;
  const ReadHandle, ErrorReadHandle: THandle; const OutputStream, ErrorStream: TStream): DWORD;
const
  BUFFER_SIZE = 512;
var
  Buffer: array[0..BUFFER_SIZE - 1] of Byte;
  BytesRead: Cardinal;
  OvLapped: OVERLAPPED;

  ErrorBuffer: array[0..BUFFER_SIZE - 1] of Byte;
  ErrorBytesRead: Cardinal;
  ErrorOvLapped: OVERLAPPED;

  HandleArray: array[0..2] of THandle;
  WaitResult: Cardinal;
begin
  ZeroMemory(@OvLapped, SizeOf(OvLapped));
  OvLapped.hEvent := CreateEvent(nil, True, False, nil);
  ZeroMemory(@ErrorOvLapped, SizeOf(ErrorOvLapped));
  ErrorOvLapped.hEvent := CreateEvent(nil, True, False, nil);
  try
    HandleArray[0] := ProcessInfo.hProcess;
    HandleArray[1] := OvLapped.hEvent;
    HandleArray[2] := ErrorOvLapped.hEvent;
    if not ReadFile(ReadHandle, Buffer, BUFFER_SIZE, BytesRead, @OvLapped)
       and (GetLastError <> ERROR_IO_PENDING) then
      RaiseLastOSError;
    if not ReadFile(ErrorReadHandle, ErrorBuffer, BUFFER_SIZE, ErrorBytesRead, @ErrorOvLapped)
       and (GetLastError <> ERROR_IO_PENDING) then
      RaiseLastOSError;

    while not Application.Terminated do
    begin
      WaitResult := MsgWaitForMultipleObjects(Length(HandleArray), HandleArray[0], False, INFINITE, QS_ALLINPUT);
      if WaitResult = WAIT_OBJECT_0 + 1 then begin

        if not GetOverlappedResult(ReadHandle, OvLapped, BytesRead, False) then
          RaiseLastOSError;
        OutputStream.Write(Buffer, BytesRead);

        ResetEvent(OvLapped.hEvent);
        if not ReadFile(ReadHandle, Buffer, BUFFER_SIZE, BytesRead, @OvLapped)
           and (GetLastError <> ERROR_IO_PENDING) then
          RaiseLastOSError;

        DoOutputProgress(ProcessInfo, OutputStream.Position);

      end else if WaitResult = WAIT_OBJECT_0 + 2 then begin

        if not GetOverlappedResult(ErrorReadHandle, ErrorOvLapped, ErrorBytesRead, False) then
          RaiseLastOSError;
        ErrorStream.Write(ErrorBuffer, ErrorBytesRead);

        ResetEvent(ErrorOvLapped.hEvent);
        if not ReadFile(ErrorReadHandle, ErrorBuffer, BUFFER_SIZE, ErrorBytesRead, @ErrorOvLapped)
           and (GetLastError <> ERROR_IO_PENDING) then
          RaiseLastOSError;

        DoErrorProgress(ProcessInfo, ErrorStream.Position);

      end else if WaitResult = WAIT_OBJECT_0 + 3 then // Message received.
        Application.ProcessMessages
      else
        break; // Process terminated.

    end;

    GetExitCodeProcess(ProcessInfo.hProcess, Result);
  finally
    CloseHandle(OvLapped.hEvent);
  end;
end;

function TAppLauncher.ExecuteA(const Path, CmdLine: AnsiString;
  out StdOut: TBytes): DWORD;
var
  lpPath, lpCmdLine: PAnsiChar;
  StartupInf: STARTUPINFOA;
  ProcessInfo: PROCESS_INFORMATION;

  WriteHandle, ReadHandle: THandle;
  dwCreationFlags: Cardinal;
  OutputStream: TBytesStream;
begin
  PreparePipe(ReadHandle, WriteHandle);
  try
    try

      StartupInf := PrepareStartupInfoA(WriteHandle);

      if CreateNoWindow then
        dwCreationFlags := CREATE_NO_WINDOW
      else
        dwCreationFlags := 0;
      if Path = '' then
        lpPath := nil
      else
        lpPath := PAnsiChar(Path);
      if CmdLine = '' then
        lpCmdLine := nil
      else
        lpCmdLine := PAnsiChar(CmdLine);
      if not CreateProcessA(lpPath, lpCmdLine, nil, nil, True, dwCreationFlags, nil, nil, StartupInf, ProcessInfo) then
        RaiseLastOSError;
      try
        try

          OutputStream := TBytesStream.Create(nil);
          try
            Result := CentralLoop(ProcessInfo, ReadHandle, OutputStream);
            StdOut := OutputStream.Bytes;
          finally
            OutputStream.Free;
          end;

        finally
          CloseHandle(ProcessInfo.hThread);
        end;
      finally
        CloseHandle(ProcessInfo.hProcess);
      end;

    finally
      CloseHandle(WriteHandle);
    end;
  finally
    CloseHandle(ReadHandle);
  end;
end;

function TAppLauncher.ExecuteA(const Path, CmdLine: AnsiString; out StdOut,
  StdErr: TBytes): DWORD;
var
  lpPath, lpCmdLine: PAnsiChar;
  StartupInf: STARTUPINFOA;
  ProcessInfo: PROCESS_INFORMATION;

  ReadHandle, WriteHandle, ErrorReadHandle, ErrorWriteHandle: THandle;
  dwCreationFlags: Cardinal;
  OutputStream: TBytesStream;
  ErrorStream: TBytesStream;
begin
  PreparePipe(ReadHandle, WriteHandle);
  try
    try
      PreparePipe(ErrorReadHandle, ErrorWriteHandle);
      try
        try
          StartupInf := PrepareStartupInfoA(WriteHandle, ErrorWriteHandle);

          if CreateNoWindow then
            dwCreationFlags := CREATE_NO_WINDOW
          else
            dwCreationFlags := 0;
          if Path = '' then
            lpPath := nil
          else
            lpPath := PAnsiChar(Path);
          if CmdLine = '' then
            lpCmdLine := nil
          else
            lpCmdLine := PAnsiChar(CmdLine);
          if not CreateProcessA(lpPath, lpCmdLine, nil, nil, True, dwCreationFlags, nil, nil, StartupInf, ProcessInfo) then
            RaiseLastOSError;
          try
            try

              OutputStream := TBytesStream.Create(nil);
              try
                ErrorStream := TBytesStream.Create(nil);
                try
                  Result := CentralLoopEx(ProcessInfo, ReadHandle, ErrorReadHandle, OutputStream, ErrorStream);
                  StdOut := OutputStream.Bytes;
                  StdErr := ErrorStream.Bytes;
                finally
                  ErrorStream.Free;
                end;
              finally
                OutputStream.Free;
              end;

            finally
              CloseHandle(ProcessInfo.hThread);
            end;
          finally
            CloseHandle(ProcessInfo.hProcess);
          end;

        finally
          CloseHandle(ErrorWriteHandle);
        end;
      finally
        CloseHandle(ErrorReadHandle);
      end;
    finally
      CloseHandle(WriteHandle);
    end;
  finally
    CloseHandle(ReadHandle);
  end;
end;

function TAppLauncher.ExecuteW(const Path, CmdLine: WideString; out StdOut: TBytes): DWORD;
var
  lpPath, lpCmdLine: PWideChar;
  StartupInf: STARTUPINFOW;
  ProcessInfo: PROCESS_INFORMATION;

  WriteHandle, ReadHandle: THandle;
  dwCreationFlags: Cardinal;
  BytesStream: TBytesStream;

  CmdLineCopy: WideString;
begin
  PreparePipe(ReadHandle, WriteHandle);
  try
    try

      StartupInf := PrepareStartupInfoW(WriteHandle);

      if CreateNoWindow then
        dwCreationFlags := CREATE_NO_WINDOW
      else
        dwCreationFlags := 0;
      if Path = '' then
        lpPath := nil
      else
        lpPath := PWideChar(Path);
      if CmdLine = '' then
        lpCmdLine := nil
      else begin
        // We need to work with a copy of CmdLine. Cf. the Microsoft
        // documentation on the CreateProcess method: "The Unicode version of
        // this function, CreateProcessW, can modify the contents of this
        // string. Therefore, this parameter cannot be a pointer to read-only
        // memory (such as a const variable or a literal string). If this
        // parameter is a constant string, the function may cause an access
        // violation.
        CmdLineCopy := CmdLine;
        UniqueString(CmdLineCopy);
        lpCmdLine := PWideChar(CmdLineCopy);
      end;
      if not CreateProcessW(lpPath, lpCmdLine, nil, nil, True, dwCreationFlags, nil, nil, StartupInf, ProcessInfo) then
        RaiseLastOSError;
      try
        try

          BytesStream := TBytesStream.Create(nil);
          try
            Result := CentralLoop(ProcessInfo, ReadHandle, BytesStream);
            StdOut := BytesStream.Bytes;
          finally
            BytesStream.Free;
          end;

        finally
          CloseHandle(ProcessInfo.hThread);
        end;
      finally
        CloseHandle(ProcessInfo.hProcess);
      end;

    finally
      CloseHandle(WriteHandle);
    end;
  finally
    CloseHandle(ReadHandle);
  end;
end;

function TAppLauncher.ExecuteW(const Path, CmdLine: WideString; out StdOut,
  StdErr: TBytes): DWORD;
var
  lpPath, lpCmdLine: PWideChar;
  StartupInf: STARTUPINFOW;
  ProcessInfo: PROCESS_INFORMATION;

  ReadHandle, WriteHandle, ErrorReadHandle, ErrorWriteHandle: THandle;
  dwCreationFlags: Cardinal;
  OutputStream: TBytesStream;
  ErrorStream: TBytesStream;

  CmdLineCopy: WideString;
begin
  PreparePipe(ReadHandle, WriteHandle);
  try
    try
      PreparePipe(ErrorReadHandle, ErrorWriteHandle);
      try
        try
          StartupInf := PrepareStartupInfoW(WriteHandle, ErrorWriteHandle);

          if CreateNoWindow then
            dwCreationFlags := CREATE_NO_WINDOW
          else
            dwCreationFlags := 0;
          if Path = '' then
            lpPath := nil
          else
            lpPath := PWideChar(Path);
          if CmdLine = '' then
            lpCmdLine := nil
          else begin
            // We need to work with a copy of CmdLine. Cf. the Microsoft
            // documentation on the CreateProcess method: "The Unicode version of
            // this function, CreateProcessW, can modify the contents of this
            // string. Therefore, this parameter cannot be a pointer to read-only
            // memory (such as a const variable or a literal string). If this
            // parameter is a constant string, the function may cause an access
            // violation.
            CmdLineCopy := CmdLine;
            UniqueString(CmdLineCopy);
            lpCmdLine := PWideChar(CmdLineCopy);
          end;
          if not CreateProcessW(lpPath, lpCmdLine, nil, nil, True, dwCreationFlags, nil, nil, StartupInf, ProcessInfo) then
            RaiseLastOSError;
          try
            try

              OutputStream := TBytesStream.Create(nil);
              try
                ErrorStream := TBytesStream.Create(nil);
                try
                  Result := CentralLoopEx(ProcessInfo, ReadHandle, ErrorReadHandle, OutputStream, ErrorStream);
                  StdOut := OutputStream.Bytes;
                  StdErr := ErrorStream.Bytes;
                finally
                  ErrorStream.Free;
                end;
              finally
                OutputStream.Free;
              end;

            finally
              CloseHandle(ProcessInfo.hThread);
            end;
          finally
            CloseHandle(ProcessInfo.hProcess);
          end;

        finally
          CloseHandle(ErrorWriteHandle);
        end;
      finally
        CloseHandle(ErrorReadHandle);
      end;
    finally
      CloseHandle(WriteHandle);
    end;
  finally
    CloseHandle(ReadHandle);
  end;
end;

end.
